#lang racket
(require redex/reduction-semantics
         "contract-nu.rkt"
         "preservation.rkt")

(test-equal (term (compat Int *)) #t)
(test-equal (term (compat * *)) #t)
(test-equal (term (compat * Bool)) #t)
(test-equal (term (compat Int Bool)) #f)
(test-equal (term (compat (∀ Y Y) (∀ X X))) #t)

(test-equal (term (contract-nu-tc mt (1 : Int ⇒ *)))
            (term *))
(test-equal (term (contract-nu-tc mt (1 : Bool ⇒ *)))
            (term illtyped))
(test-equal (term (contract-nu-tc mt (1 : Int ⇒ pp Int)))
            (term Int))
(test-equal (term (contract-nu-tc mt (1 : Int ⇒ pp Bool)))
            (term illtyped))
(test-equal (term (contract-nu-tc mt (blame pp Int)))
            (term Int))
(test-equal (term (contract-nu-tc mt (Λ X (Λ Y (λ x : X (x : X ⇒ pp Y))))))
            (term illtyped))
(test-equal (term (contract-nu-tc mt (ν X := Int (1 : Int ⇒ pp X))))
            (term illtyped))
(test-equal (term (contract-nu-tc mt (ν X := Int (λ x : X x))))
            (term (Int -> Int)))
(test-equal (term (contract-nu-tc mt (ν X := * ((3 : Int ⇒ *) : X ⇒ *))))
            (term *))

;; basic blame calculus reduction tests
; Wrap
(test--> contract-nu-red
         (term ((λ yy : Int false) : (Int -> Bool) ⇒ pp (Int -> Bool)))
         (term (λ x : Int (((λ yy : Int false)
                            (x : Int ⇒ (bar pp) Int))
                           : Bool ⇒ pp Bool))))
; Wrap
(test--> contract-nu-red
         (term ((λ yy : Int false) : (Int -> Bool) ⇒ pp (* -> *)))
         (term (λ x : * (((λ yy : Int false)
                          (x : * ⇒ (bar pp) Int))
                         : Bool ⇒ pp *))))
; Id
(test--> contract-nu-red
         (term (1 : Int ⇒ pp Int))
         (term 1))
; Id
(test--> contract-nu-red
         (term (1 : X ⇒ pp X))
         (term 1))

; Ground
(test--> contract-nu-red
         (term (1 : Int ⇒ pp *))
         (term ((1 : Int ⇒ pp Int) : Int ⇒ *)))
; Ground
(test--> contract-nu-red
         (term (1 : Bool ⇒ pp *))
         (term ((1 : Bool ⇒ pp Bool) : Bool ⇒ *)))
; Ground
(test--> contract-nu-red
         (term (1 : X ⇒ pp *))
         (term ((1 : X ⇒ pp X) : X ⇒ *)))
; Ground
(test--> contract-nu-red
         (term ((λ xx : Int xx) : (Int -> Int) ⇒ pp *))
         (term (((λ xx : Int xx) : (Int -> Int) ⇒ pp (* -> *)) : (* -> *) ⇒ *)))
; Ground
(test--> contract-nu-red
         (term ((Λ X 1) : (∀ X Int) ⇒ pp *))
         (term (((Λ X 1) : (∀ X Int) ⇒ pp Int) : Int ⇒ *)))
; Collapse
(test--> contract-nu-red
         (term ((1 : Int ⇒ *) : * ⇒ pp Int))
         (term (1 : Int ⇒ pp Int)))
; Conflict
(test--> contract-nu-red
         (term ((1 : Int ⇒ *) : * ⇒ pp Bool))
         (term (blame pp Bool)))
; Abort
(test--> contract-nu-red
         (term (+ ((λ x : Bool 1) (blame p Bool)) 1))
         (term (blame p Int)))

; IsTrue
(test--> contract-nu-red
         (term ((1 : Int ⇒ *) is Int))
         (term true))
; IsFalse
(test--> contract-nu-red
         (term ((1 : Int ⇒ *) is Bool))
         (term false))
; IsSeal 
(test--> contract-nu-red
         (term ((1 : X ⇒ *) is Int))
         (term (blame Is Bool)))

; TyBeta
(test--> contract-nu-red
         (term ((Λ X (λ x : X x)) Int))
         (term (ν X := Int (λ x : X x))))
; NuConst
(test--> contract-nu-red
         (term (ν X := Int 1))
         (term 1))
; NuWrap
(test--> contract-nu-red
         (term (ν X := Int (λ x : Int x)))
         (term (λ x : Int (ν X := Int x))))
; NuWrap
(test--> contract-nu-red
         (term (ν X := Int (λ x : X x)))
         (term (λ x : Int (ν X := Int x))))
; NuTyWrap
(test--> contract-nu-red
         (term (ν X := (Int -> Z) (Λ Y (λ y : Y y))))
         (term (Λ Y (ν X := (Int -> Z) (λ y1 : Y y1)))))
; NuTyWrap
(test--> contract-nu-red
         (term (ν X := (Int -> Z) (Λ Z (λ z : Z z))))
         (term (Λ Z1 (ν X := (Int -> Z) (λ z1 : Z1 z1)))))
; NuTyWrap
(test--> contract-nu-red
         (term (ν X := (Int -> Z) (Λ X (λ x : X x))))
         (term (Λ X1 (ν X := (Int -> Z) (λ x1 : X1 x1)))))

; NuGround
(test--> contract-nu-red
         (term (ν X := Int (1 : Int ⇒ *)))
         (term ((ν X := Int 1) : Int ⇒ *)))
; NuGround
(test--> contract-nu-red
         (term (ν X := Int (1 : Y ⇒ *)))
         (term ((ν X := Int 1) : Y ⇒ *)))

; NuTamper
(test--> contract-nu-red
         (term (ν X := Int (1 : X ⇒ *)))
         (term (blame ν *)))


; Instantiate 
(test--> contract-nu-red
         (term ((Λ X (λ x : X x)) : (∀ X (X -> X)) ⇒ pp (* -> *)))
         (term (((Λ X (λ x : X x)) *) : (* -> *) ⇒ pp (* -> *))))
; Generalize
(test--> contract-nu-red
         (term (1 : Int ⇒ p (∀ X Int)))
         (term (Λ X (1 : Int ⇒ p Int))))
; Generalize
(test--> contract-nu-red
         (term ((λ x : * x) : (* -> *) ⇒ p (∀ X (X -> X))))
         (term (Λ X ((λ x : * x) : (* -> *) ⇒ p (X -> X)))))
; Generalize (and importantly, not Instantiate)
(test--> contract-nu-red
         (term ((Λ X (λ x : X x)) : (∀ X (X -> X)) ⇒ p2 (∀ X (X -> *))))
         (term (Λ X ((Λ X (λ x : X x)) : (∀ X (X -> X)) ⇒ p2 (X -> *)))))
; Conflict
(test--> contract-nu-red
         (term ((1 : X ⇒ *) : * ⇒ qq Int))
         (term (blame qq Int)))
; Conflict
(test--> contract-nu-red
         (term ((1 : X ⇒ *) : * ⇒ qq Y))
         (term (blame qq Y)))
; IsSeal
(test--> contract-nu-red
         (term ((1 : X ⇒ *) is Int))
         (term (blame Is Bool)))
; IsSeal
(test--> contract-nu-red
         (term ((1 : X ⇒ *) is Int))
         (term (blame Is Bool)))
; Ground
(test--> contract-nu-red
         (term (ν X := Int (1 : X ⇒ (bar pp) *)))
         (term (ν X := Int ((1 : X ⇒ (bar pp) X) : X ⇒ *))))
; Ground
(test--> contract-nu-red
         (term ((1 : Int ⇒ *) : * ⇒ qq *))
         (term (1 : Int ⇒ qq *)))

; test values (some of which dont' typecheck...)
(test--> contract-nu-red
         (term (1 : X ⇒ *)))
(test--> contract-nu-red
         (term (1 : Int ⇒ *)))
(test--> contract-nu-red
         (term (Λ X (1 : X ⇒ *))))
(test--> contract-nu-red
         (term 1))
 
(test--> contract-nu-red
         (term ((λ x : Bool x) ((1 : Int ⇒ *) is Bool)))
         (term ((λ x : Bool x) false)))

(test--> contract-nu-red
         (term ((λ x : Bool x) ((1 : Int ⇒ *) : * ⇒ pp Bool)))
         (term ((λ x : Bool x) (blame pp Bool))))
(test-->> contract-nu-red
          (term ((λ x : Bool x) ((1 : Int ⇒ *) : * ⇒ pp Bool)))
          (term (blame pp Bool)))

(test-equal (term (<: Int Int)) #t)
(test-equal (term (<: Int *)) #t)
(test-equal (term (<: Bool *)) #t)
(test-equal (term (<: * Int)) #f)
(test-equal (term (<: * Bool)) #f)
(test-equal (term (<: (* -> *) *)) #t)
(test-equal (term (<: (Int -> Int) (* -> *))) #f)
(test-equal (term (<: (Int -> Int) (Int -> *))) #t)
(test-equal (term (<: (Int -> Int) (* -> Int))) #f)
(test-equal (term (<: (* -> Int) (* -> *))) #t)
(test-equal (term (<: (Int -> *) (* -> *))) #f)
(test-equal (term (<: (∀ X (X -> X)) (Int -> Int))) #t)
(test-equal (term (<: (∀ X (X -> X)) (* -> *))) #t)
(test-equal (term (<: Int (∀ X Int))) #t)
 
(test-equal (term (sf<: (1 : Int ⇒ pp *) pp))
            #t)
(test-equal (term (sf<: (1 : (* -> *) ⇒ pp (* -> *)) pp))
            #t)
(test-equal (term (sf<: (1 : (Int -> *) ⇒ pp (* -> *)) pp))
            #f)
(test-equal (term (sf<: (1 : (* -> *) ⇒ pp (* -> Int)) pp))
            #f)
(test-equal (term (sf<: (1 : (* -> *) ⇒ qq (* -> Int)) pp))
            #t)
(let ([saf (term (1 : (* -> *) ⇒ qq (Int -> Int)))]
      [uns (term (1 : (* -> *) ⇒ pp (Int -> Int)))])
  (test-equal (term (sf<: (,uns : (Int -> Int) ⇒ rr *) pp)) #f)
  (test-equal (term (sf<: (,saf : (Int -> Int) ⇒ rr *) pp)) #t)
  (test-equal (term (sf<: ((,uns : (Int -> Int) ⇒ rr (* -> *)) : (* -> *) ⇒ *) pp)) #f)
  (test-equal (term (sf<: ((,saf : (Int -> Int) ⇒ rr (* -> *)) : (* -> *) ⇒ *) pp)) #t)
  (test-equal (term (sf<: ((,uns : (Int -> Int) ⇒ rr *)  is Bool) pp)) #f)
  (test-equal (term (sf<: ((,saf : (Int -> Int) ⇒ rr *)  is Bool) pp)) #t)
  (test-equal (term (sf<: (blame pp Int) pp)) #f)
  (test-equal (term (sf<: (blame (bar pp) Int) pp)) #f)
  (test-equal (term (sf<: (blame pp Int) (bar pp))) #f)
  (test-equal (term (sf<: (blame qq Int) pp)) #t)
  (test-equal (term (sf<: (Λ X ,uns) pp)) #f)
  (test-equal (term (sf<: (Λ X ,saf) pp)) #t)
  (test-equal (term (sf<: ((Λ X ,uns) Int) pp)) #f)
  (test-equal (term (sf<: ((Λ X ,saf) Int) pp)) #t)
  (test-equal (term (sf<: (ν X := Int ,uns) pp)) #f)
  (test-equal (term (sf<: (ν X := Int ,saf) pp)) #t)
  (test-equal (term (sf<: 1 pp)) #t)
  (test-equal (term (sf<: (+ (,uns 1) 1) pp)) #f)
  (test-equal (term (sf<: (+ (,saf 1) 1) pp)) #t)
  (test-equal (term (sf<: xx pp)) #t)
  (test-equal (term (sf<: (λ x : Int ,uns) pp)) #f)
  (test-equal (term (sf<: (λ x : Int ,saf) pp)) #t)
  (test-equal (term (sf<: (,uns 1) pp)) #f)
  (test-equal (term (sf<: (,saf 1) pp)) #t)
  (test-equal (term (sf<: ((λ x : (Int -> Int) x) ,uns) pp)) #f)
  (test-equal (term (sf<: ((λ x : (Int -> Int) x) ,saf) pp)) #t)
  (test-equal (term (sf<: 1 ν)) #f)
  (test-equal (term (sf<: true Is)) #f))

(test-results)

(parameterize ([current-tc (λ (x) (term (contract-nu-tc mt ,x)))]
               [current-red contract-nu-red]
               [current-same-ty (λ (x y) (term (same-ty ,x ,y)))])
  (type-preservation (term (((λ x : * x) : (* -> *) ⇒ pp (Int -> Int)) 1))
                     (term 1))
  (type-preservation (term ((((λ x : * x) : (* -> *) ⇒ pp *) : * ⇒ pp (Int -> Int)) 1))
                     (term 1))
  (type-preservation (term ((λ f : (* -> *) (f (f ((+ 1 2) : Int ⇒ qq *))))
                            ((λ x : * (x is Bool)) : (* -> Bool) ⇒ pp (* -> *))))
                     (term (true : Bool ⇒ *)))
  
  (type-preservation (term ((((λ x : * x) : (* -> *) ⇒ pp (∀ X (X -> X)))
                             Int)
                            1))
                     (term 1))
  
  (type-preservation (term (((Λ X (λ x : X x)) 
                             :
                             (∀ X (X -> X))
                             ⇒
                             pp
                             (Int -> Int))
                            3))
                     (term 3))
  
  (type-preservation (term ((((λ x : * x) : (* -> *) ⇒ pp (∀ X (X -> X))) Int) 2))
                     (term 2))
  
  (type-preservation (term (((Λ X ((Λ Y (λ x : Y x)) X))
                             Int)
                            1))
                     (term 1))
  
  (type-preservation (term (((Λ Y ((Λ X (λ y : Y y)) Y))
                             Int)
                            1))
                     (term 1))
  
  (type-preservation
   (term (((Λ X (λ x : X ((x : X ⇒ qq *) : * ⇒ pp *)))
           Int)
          1))
   (term (blame ν *)))
    
  (type-preservation
   (term 
    (((Λ X (λ x : X (((x : X ⇒ p2 *) : * ⇒ p3 *) is Bool)))
      Int)
     1))
   (term (blame Is Bool)))
  
  (type-preservation (term ((((λ x : * ((λ y : Bool x) (x is Int))) : (* -> *) ⇒ (bar p2) *) : * ⇒ p3 (Int -> Int)) 3))
                     (term 3))
  
  (type-preservation (term ((((λ x : * ((λ y : Bool x) (x is Int))) : (* -> *) ⇒ (bar p2) (∀ X (X -> X))) : (∀ X (X -> X)) ⇒ pp (Int -> Int)) 3))
                     (term (blame Is Int)))
  
  (type-preservation (term (((Λ X (λ x : X ((x : X ⇒ qq *) : * ⇒ pp *)))
                             Int)
                            1))
                     (term (blame ν *)))
  
  (type-preservation (term (((Λ X (λ x : X (x : X ⇒ pp X))) Int) 1))
                     (term 1))
  (type-preservation (term (((Λ X (λ x : X ((x : X ⇒ qq *) : * ⇒ pp X))) Int) 1))
                     (term 1))
  
  (type-preservation (term ((λ x : * (x : * ⇒ p1 Int))
                            (true : Bool ⇒ p2 *)))
                     (term (blame p1 Int)))
  
  (type-preservation (term (((((Λ X (Λ Y (λ x : X (λ y : Y x)))) Int) Bool) 1) true))
                     (term 1))
  
  ;; this is the same as a starred [*] example below
  (type-preservation (term ((((Λ X (λ x : X x)) : (∀ X (X -> X)) ⇒ p2 (∀ X (X -> *))) Int) 2))
                     (term (blame ν *)))
  
  (type-preservation (term ((((λ x : * x) : (* -> *) ⇒ ppp (∀ X (X -> *))) Int) 3))
                     (term (blame ν *)))
  
  (show-steps contract-nu-red)
  
  (parameterize ([current-sf (λ (x) (term (sf<: ,x pp)))])
    (sf<:-preservation (term ((λ x : Int x) 1)))
    (sf<:-preservation (term (((λ x : * 1) : (* -> Int) ⇒ pp (Int -> *)) 1)))
    (sf<:-preservation (term (((λ x : Int x) : (Int -> Int) ⇒ pp (Int -> Int)) 1)))
    (sf<:-preservation (term ((((λ x : Int x) : (Int -> Int) ⇒ pp (∀ X (Int -> Int))) Bool) 1)))
    
    (sf<:-preservation (term ((((λ x : * x) : (* -> *) ⇒ pp (∀ X (* -> *))) Bool) (1 : Int ⇒ qq *)))) 
    
    (sf<:-preservation
     (term (((Λ X (λ x : X ((x : X ⇒ qq *) : * ⇒ pp *))) Int) 1)))
    
    (sf<:-preservation
     (term 
      (((Λ X (λ x : X (((x : X ⇒ qq *) : * ⇒ pp *) is Bool)))
        Int)
       1)))
    
    ;; this is the same as the starred [*] example above 
    (sf<:-preservation
     (term ((((Λ X (λ x : X x)) : (∀ X (X -> X)) ⇒ p2 (∀ X (X -> *))) Int) 2)))
    
    (sf<:-preservation
     (term (((λ x : Bool (λ y : Bool (+ 1 2))) ((1 : Int ⇒ *) is Bool)) ((2 : Int ⇒ *) is Int))))
    
    (sf<:-preservation
     (term ((1 : Int ⇒ p *) : * ⇒ q Bool)))

    (sf<:-preservation (term (((((Λ X (Λ Y (λ x : X (λ y : Y x)))) Int) Bool) 1) true)))
    
    (show-steps contract-nu-red)))
