#lang racket
(require redex/reduction-semantics
         "nu.rkt"
         "same-ty.rkt"
         "subst.rkt")
(provide contract-nu 
         contract-nu-tc
         contract-nu-red
         compat
         sf<:
         <:)

(define-extended-language contract-nu
  nu
  ((A B C) .... *)
  ((G H) ι (* -> *) X)
  ((p q) x (bar x))
  (l p Is ν)
  ((s t) c (op t ...) x (λ x : A t) (t s)
         (let x = s in t) (s : A ⇒ p B) (s : G ⇒ *)
         (s is G) (blame l A) (Λ X t) (t A) (ν X := A t))
  ((v w) c (λ x : A t) (v : G ⇒ *) (Λ X v))
  (E hole (op v ... E t ...) (E s) (v E)
     (let x = E in t) (E : A ⇒ p B) (E : G ⇒ *)
     (E is G) (Λ X E) (E A) (ν X := A E))
  (G/none G none))

(define-metafunction/extension nu-tc contract-nu
  contract-nu-tc : Γ t -> A/illtyped
  
  [(contract-nu-tc Γ (t : A_1 ⇒ p B))
   B
   (where A_2 (contract-nu-tc Γ t))
   (side-condition (term (same-ty/Γ Γ A_1 A_2)))
   
   ;; not sure if this should be A_1 or A_2
   ;; (or something else)
   (side-condition (term (compat A_1 B)))]
  
  [(contract-nu-tc Γ (t : G ⇒ *))
   *
   (where A (contract-nu-tc Γ t))
   (side-condition (term (same-ty/Γ Γ G A)))]
   
  [(contract-nu-tc Γ (blame l A))
   A]
  
  
  ;; from the poly blame calculus figure
  [(contract-nu-tc Γ (s is G))
   Bool
   (where * (contract-nu-tc Γ s))
   (side-condition (term (not-X G)))]
  [(contract-nu-tc Γ (Λ X t))
   (∀ Z B)
   (where Z ,(variable-not-in (term Γ) (term X)))
   (where B (contract-nu-tc (Γ Z) (subst X Z t)))]
  [(contract-nu-tc Γ (ν X := A t))
   (subst Z A B)
   (side-condition (term (wft Γ A)))
   (where Z ,(variable-not-in (term Γ) (term X)))
   (where B (contract-nu-tc (Γ X A) (subst X Z t)))])

(define-relation contract-nu
  compat ⊂ A × A
  [(compat A B) (same-ty A B)]
  [(compat A *)]
  [(compat * A)]
  [(compat (A -> B) (A_2 -> B_2))
   (compat A_2 A)
   (compat B B_2)]
  [(compat (∀ X A) B)
   (compat (subst X * A) B)]
  [(compat A (∀ X B))
   (compat A B)
   (not-in X (ftv A))])

(define contract-nu-red
  (extend-reduction-relation
   nu-red
   contract-nu
   
   ;; rules for stlc+contracts

   (---> (v : (A -> B) ⇒ p (A_2 -> B_2))
         (λ x : A_2 ((v (x : A_2 ⇒ (inv p) A)) : B ⇒ p B_2))
         "Wrap")
   
   
   (---> (v : G ⇒ p G)
         v
         (side-condition (term (different G (* -> *))))
         "Id")
   (---> (v : A ⇒ p *)
         ((v : A ⇒ p G) : G ⇒ *)
         "Ground"
         (side-condition (term (different * A)))
         (where G (find-compat A)))
   (---> ((v : G ⇒ *) : * ⇒ p A)
         (v : G ⇒ p A)
         "Collapse"
         (side-condition (term (compat G A))))
   (---> ((v : G ⇒ *) : * ⇒ p A)
         (blame p A)
         "Conflict"
         (side-condition (not (term (compat G A)))))
   
   ;; this rule does not preserve types in a simple way. 
   ;; Specifically, if the rule were just:
   ;;   E[blame p A] -> blame p A
   ;; then the type on the left might not be A, but the type
   ;; on the right is supposed to be A. So, we just call
   ;; the type checker here ... (yikes)
   ;; note that this rule will be stuck if the left-hand
   ;; side of the rule does not typecheck (not that that should
   ;; happen, mind you)
   (--> (in-hole E (blame l A))
        (blame l B)
        (side-condition (term (different E hole)))
        (where B (contract-nu-tc mt (in-hole E (blame l A))))
        "Abort")
   
   ;; rules for parametric contracts
   
   (---> ((v : G ⇒ *) is G)
         true
         "IsTrue" 
         (side-condition (term (not-X G))))
   (---> ((v : H ⇒ *) is G)
         false
         "IsFalse"
         (side-condition (term (not-X H)))
         (side-condition (term (different G H))))
   (---> ((v : X ⇒ *) is G)
         (blame Is Bool)
         "IsSeal")
   
   ;; these four rules replace the ones with the same names from nu.rkt
   (---> ((Λ X v) A) 
         (ν X := A v)
         "TyBeta")
   (---> (ν X := A c) 
         c 
         "NuConst")
   (---> (ν X := A (λ y : B t))
         (λ y : (subst X A B) (ν X := A t))
         "NuWrap")
   (---> (ν X := A (Λ Y v))
         (Λ Z (ν X := A (subst Y Z v)))
         (where Z ,(if (and (term (different Y X))
                            (term (not-in Y (ftv A))))
                       (term Y)
                       (variable-not-in (term (X A)) (term Y))))
         "NuTyWrap")
   
   (---> (ν X := A (v : G ⇒ *))
         ((ν X := A v) : G ⇒ *)
         (side-condition (term (different G X)))
         "NuGround")
   (---> (ν X := A (v : X ⇒ *))
         (blame ν *)
         "NuTamper")
   
   (---> (v : A ⇒ p (∀ X B))
         (Λ X (v : A ⇒ p B))
         "Generalize")
   (---> (v : (∀ X A) ⇒ p B)
         ((v *) : (subst X * A) ⇒ p B)
         "Instantiate"
         (side-condition (term (not-∀ B)))
         (side-condition (term (different B *))))
   with
   [(--> (in-hole E a) (in-hole E b)) (---> a b)]))

(define-metafunction contract-nu
  find-compat : A -> G/none
  [(find-compat (A -> B)) (* -> *)]
  [(find-compat Int) Int]
  [(find-compat Bool) Bool]
  [(find-compat X) X]
  [(find-compat (∀ X A)) (find-compat (subst X * A))]
  [(find-compat A) none])

(define-metafunction contract-nu
  not-X : any -> any
  [(not-X X) #f]
  [(not-X any) #t])

(define-metafunction contract-nu
  not-∀ : any -> any
  [(not-∀ (∀ X A)) #f]
  [(not-∀ any) #t])

(define-relation contract-nu
  <: ⊂ A × A
  [(<: A A)]
  [(<: A *)
   (<: A Int)]
  [(<: A *)
   (<: A Bool)]
  [(<: A *)
   (<: A (* -> *))]
  [(<: X *)] ;; is this sufficiently general?
  
  [(<: (A -> B) (A_2 -> B_2))
   (<: A_2 A)
   (<: B B_2)]
  [(<: A (∀ X B))
   (not-in X (ftv A))
   (<: A B)]
  ;; these rules are specific uses of the remaining ("witness") rule;
  ;; but not all of the possible ones
  [(<: (∀ X A) B)
   (<: (subst X * A) B)]
  [(<: (∀ X A) B)
   (<: (subst X Int A) B)]
  [(<: (∀ X A) B)
   (<: (subst X Bool A) B)]
  [(<: (∀ X A) B)
   (<: (subst X * A) B)])

(define-relation contract-nu
  sf<: ⊂ t × l
  [(sf<: (s : A ⇒ p B) p)
   (<: A B)
   (sf<: s p)]
  [(sf<: (s : A ⇒ q B) p)
   ,(equal? (term (inv p)) (term q))
   (<: A B)
   (sf<: s p)]
  [(sf<: (s : A ⇒ q B) p)
   (different q p)
   (different q (inv p))
   (sf<: s p)]
  [(sf<: (s : G ⇒ *) p)
   (sf<: s p)]
  [(sf<: (s is G) p)
   (sf<: s p)]
  [(sf<: (blame q A) p)
   (different q p)
   (different q (inv p))]
  [(sf<: (blame Is A) p)]
  [(sf<: (blame ν A) p)]
  [(sf<: c p)]
  [(sf<: (op t ...) p)
   (all-sf<: t ... p)]
  [(sf<: x p)]
  [(sf<: (λ x : A t) p)
   (sf<: t p)]
  [(sf<: (t s) p)
   (sf<: t p)
   (sf<: s p)]
  [(sf<: (Λ X t) p)
   (sf<: t p)]
  [(sf<: (t A) p)
   (sf<: t p)]
  [(sf<: (ν X := A t) p)
   (sf<: t p)])

(define-relation contract-nu
  [(all-sf<: l)]
  [(all-sf<: t_1 t_2 ... l)
   (sf<: t_1 l)
   (all-sf<: t_2 ... l)])

;; give this one a lax contract since there is no inline 'or' patterns
(define-metafunction contract-nu
  inv : p -> p
  [(inv (bar p)) p]
  [(inv p) (bar p)])
