#lang racket
(require redex/reduction-semantics
         "nu.rkt"
         "preservation.rkt"
         "same-ty.rkt")
(provide nu-red-test
         nu-preservation)


;; basic tests for each reduction relation case
(define (nu-red-test nu-red)
  (test--> nu-red
           (term ((Λ Y (λ y : Y y)) Int))
           (term (ν Y := Int (λ y : Y y))))
  (test--> nu-red
           (term (ν Y := Int 1))
           (term 1))
  (test--> nu-red
           (term (ν Y := Int (λ y : Bool y)))
           (term (λ y : Bool (ν Y := Int y))))
  (test--> nu-red
           (term (ν XX := Int (Λ YY (λ yy : Int yy))))
           (term (Λ YY (ν XX := Int  (λ yy : Int yy)))))
  ;; misc slightly bigger tests
  
  (test-->> nu-red
            (term (((Λ X (λ x : X x))
                    Int)
                   1))
            (term 1))
  
  (test-->> nu-red
            (term ((((Λ X (λ f : (X -> X) (λ x : X (f x))))
                     Int)
                    (λ y : Int (+ y 1)))
                   3))
            (term 4))
  
  (test-->> nu-red
            (term ((((Λ X (λ f : (X -> X) (λ x : X (f (f x)))))
                     Int)
                    (λ y : Int (+ y y)))
                   3))
            (term 12))
  
  (test-->> nu-red
            (term ((((Λ X (Λ Y (λ y : Y y)))
                     Int)
                    Bool)
                   true))
            (term true))
  
  (test-->> nu-red
            (term (((((Λ X (Λ Y (λ f : (X -> Y) (λ x : X (f x)))))
                      Int)
                     Bool)
                    (λ y : Int true))
                   3))
            (term true)))

(nu-red-test nu-red)

;; type checking tests
(test-equal (term (nu-tc mt (Λ X (λ x : X x))))
            (term (∀ X (X -> X))))
(test-equal (term (nu-tc mt (λ x : Int (Λ X (λ y : Int x)))))
            (term (Int -> (∀ X (Int -> Int)))))
(test-equal (term (nu-tc mt ((Λ X (λ x : X x)) Int)))
            (term (Int -> Int)))
(test-equal (term (nu-tc mt ((Λ Y (Λ X (λ x : X (λ y : Y x))))
                                   Int)))
            (term (∀ X1 (X1 -> (Int -> X1)))))
(test-equal (term (nu-tc mt (ν XX := Int true)))
            (term Bool))
(test-equal (term (nu-tc mt (ν XX := Int (λ xx : XX true))))
            (term (Int -> Bool)))

(test-equal (term (nu-tc mt ((ν X := Int (λ x : X x)) 1)))
            (term Int))

(test-equal (term (nu-tc mt (ν X := Int (λ x : X (+ x 1)))))
            (term (Int -> Int)))
(test-equal (term (nu-tc mt ((Λ X (λ x : X (+ x 1))) Int)))
            (term illtyped))


(test-results)

(define (nu-preservation)
  (type-preservation (term 1) (term 1))
  (type-preservation (term ((λ x : Int x) 1)) (term 1))
  
  (type-preservation (term (((Λ X (λ x : X x))
                             Int)
                            1))
                     (term 1))
  
  (type-preservation (term ((((Λ X (λ f : (X -> X) (λ x : X (f x))))
                              Int)
                             (λ y : Int (+ y 1)))
                            3))
                     (term 4))
  
  (type-preservation (term ((((Λ X (λ f : (X -> X) (λ x : X (f (f x)))))
                              Int)
                             (λ y : Int (+ y y)))
                            3))
                     (term 12))
  
  (type-preservation (term ((((Λ X (Λ Y (λ y : Y y)))
                              Int)
                             Bool)
                            true))
                     (term true))
  
  (type-preservation (term (((((Λ X (Λ Y (λ f : (X -> Y) (λ x : X (f x)))))
                               Int)
                              Bool)
                             (λ y : Int true))
                            3))
                     (term true))
  
  (type-preservation (term ((ν X := Int (λ x : X (+ x 1))) 1))
                     (term 2)))

(parameterize ([current-tc (λ (x) (term (nu-tc mt ,x)))]
               [current-red nu-red]
               [current-same-ty (λ (x y) (term (same-ty ,x ,y)))])
  (nu-preservation)
  (show-steps nu-red))
