#lang racket
(require redex/reduction-semantics
         "stlc.rkt"
         "subst.rkt"
         "same-ty.rkt")

(provide nu nu-red nu-tc 
         wft
         different)

(define-extended-language nu
  stlc
  ((A B C) .... X (∀ X B))
  ((s t) .... 
         (Λ X t)
         (t A) 
         (ν X := A t))
  (Γ .... (Γ X) (Γ X A))
  ((v w) .... (Λ X v))
  (E .... 
     (Λ X E) 
     (E A)
     (ν X := A E))
     
  (X 
   (side-condition (name X variable-not-otherwise-mentioned)
                   (regexp-match #rx"^[A-Z]" (symbol->string (term X)))))
  (Y X)
  (Z X))

(define nu-red
  (extend-reduction-relation
   stlc-red
   nu
   (---> ((Λ X v) A) 
         (ν X := A v)
         "TyBeta")
   (---> (ν X := A c) 
         c 
         "NuConst")
   (---> (ν X := A (λ y : B t))
         (λ y : (subst X A B) (ν X := A t))
         "NuWrap")
   (---> (ν X := A (Λ Y v))
         (Λ Y (ν X := A v))
         (side-condition (term (different Y X)))
         (side-condition (term (not-in Y (ftv A))))
         "NuTyWrap")
   with
   [(--> (in-hole E a) (in-hole E b)) (---> a b)]))

(define-metafunction/extension stlc-tc nu
  nu-tc : Γ t -> A/illtyped
  [(nu-tc Γ (Λ X t))
   (∀ X B)
   (where B (nu-tc (Γ X) t))]
  [(nu-tc Γ (t A))
   (subst X A B)
   (where (∀ X B) (nu-tc Γ t))
   (side-condition (term (wft Γ A)))]
  [(nu-tc Γ (ν X := A t))
   (subst X A B)
   (where B (nu-tc (Γ X A) t))
   (side-condition (term (wft Γ A)))]
  [(nu-tc Γ (t : B ⇒ X C))
   C
   (where A (lookup X Γ))
   (where B (nu-tc Γ t))
   (where C (subst X A B))]
  [(nu-tc Γ (t : C ⇒ (bar X) B))
   B
   (where A (lookup X Γ))
   (where C (nu-tc Γ t))
   (where C (subst X A B))]
  
  ;; additional variable cases to cover 
  ;; the extended Γ relation
  [(nu-tc (Γ X A) x)
   (nu-tc Γ x)]
  [(nu-tc (Γ X) x)
   (nu-tc Γ x)])

(define-metafunction nu
  different : any any -> any
  [(different any_1 any_1) #f]
  [(different any_1 any_2) #t])

