#lang racket
(require redex/reduction-semantics)
(provide current-tc current-red current-same-ty current-sf
         show-steps
         type-preservation
         sf<:-preservation)

(define preservation-steps 0)
(define preservation-step-names (make-hash))

(define current-tc (make-parameter 'unset-current-tc))
(define current-red (make-parameter 'unset-current-red))
(define current-same-ty (make-parameter 'unset-current-same-ty))
(define current-sf (make-parameter 'unset-current-sf))

(define (preservation orig-t prop fail)
  (let loop ([t orig-t]
             [i 1])
    (define next (apply-reduction-relation/tag-with-names (current-red) t))
    (cond
      [(null? next) t]
      [(null? (cdr next))
       (define step-name (string->symbol (list-ref (car next) 0)))
       (define next-term (list-ref (car next) 1))
       (unless (prop next-term)
         (fail 
          'preservation
          "step ~a does not preserve invariant:\n original-term: ~s\n     this-term: ~s"
          i orig-t next-term))
       (set! preservation-steps (+ preservation-steps 1))
       (hash-set! preservation-step-names step-name (+ (hash-ref preservation-step-names step-name 0) 1))
       (loop next-term (+ i 1))]
      [else
       (fail 'preservation 
              "this term (step ~a) has multiple possibilites:\n  ~s" 
              i
              t)])))

(define (type-preservation orig-t expected)
  (let/ec k
    (define (fail sym . args) 
      (fprintf (current-error-port) "~a: ~a\n" sym (apply format args))
      (k (void)))
    (define t-ty ((current-tc) orig-t))
    (define res-ty ((current-tc) expected))
    (when (eq? (term illtyped) t-ty)
      (fail 'type-preservation "the term ~s is illtyped" orig-t))
    (when (eq? (term illtyped) res-ty)
      (fail 'type-preservation "the result ~s is illtyped" expected))
    (unless (term (same-ty ,res-ty ,t-ty))
      (fail
       'type-preservation
       "the term and the result have different types:\n term ~s\n  res ~s\n"
       t-ty res-ty))
    (define got
      (preservation orig-t 
                    (λ (x) 
                      (define next-ty ((current-tc) x))
                      (and (not (eq? next-ty 'illtyped))
                           ((current-same-ty) t-ty next-ty)))
                    fail))
    (unless (equal? expected got)
      (fail 'type-preservation "reduction sequence did not end with the expected:\n     ~s\ngot: ~s" expected got))))

(define (sf<:-preservation orig-t) 
  (let/ec k
    (define (fail sym . args) 
      (fprintf (current-error-port) "~a: ~a\n" sym (apply format args))
      (k (void)))
    (unless ((current-sf) orig-t)
      (fail 'sf<:-preservation "the original input is not sf:\n  ~s" orig-t))
    (preservation orig-t (current-sf) fail)))
   
(define (steps red term)
  (let loop ([term term])
    (let ([nexts (apply-reduction-relation red term)])
      (cond
        [(null? nexts) (list term)]
        [else (cons term (loop (car nexts)))]))))

(define (show-steps red)
  (let ([missing '()])
    (for ([x (in-list (reduction-relation->rule-names red))])
      (unless (hash-ref preservation-step-names x #f)
        (unless (member x missing)
          (set! missing (cons x missing)))))
    (set! missing (sort missing string<=? #:key symbol->string))
    (printf "preservation checked for ~a example steps~a\n" 
            preservation-steps
            (if (null? missing)
                ", using all of the reduction rules"
                (format ", but no reduction sequence used the rules: ~s" missing)))
    (set! preservation-steps 0)
    (set! preservation-step-names (make-hash))))
