#lang racket
(require redex/reduction-semantics
         "same-ty.rkt"
         "subst.rkt")

(provide stlc stlc-red stlc-tc
         tyop tyc)

(define-language stlc
  (ι Int Bool)
  ((A B C) ι (A -> B))
  ((s t) c (op t ...) x (λ x : A t) (t s))
  (Γ mt (Γ x A))
  ((v w) c (λ x : A t))
  (E hole (E s) (v E) (op v ... E t ...))
  (op + - = if)
  (c number true false)
  ((x y z) 
   (side-condition (name x variable-not-otherwise-mentioned)
                   (regexp-match #rx"^[a-z]" (symbol->string (term x)))))
  
  ;; used only as the range of the tc metafunction
  (A/illtyped A illtyped))

(define stlc-red
  (reduction-relation 
   stlc
   (---> ((λ x : A t) v)
         (subst x v t)
         "βv")
   (---> (op c ...)
         (δ op c ...)
         "δ")
   with
   [(--> (in-hole E a) (in-hole E b)) (---> a b)]))

(define-metafunction stlc 
  δ : op c ... -> v
  [(δ + number_1 number_2) ,(+ (term number_1) (term number_2))]
  [(δ - number_1 number_2) ,(- (term number_1) (term number_2))]
  [(δ = number_1 number_1) true]
  [(δ = number_1 number_2) false]
  [(δ if true) (λ x : (Int -> Int) (λ y : (Int -> Int) (x 0)))]
  [(δ if false) (λ x : (Int -> Int) (λ y : (Int -> Int) (y 0)))])
  
(define-metafunction stlc
  stlc-tc : Γ t -> A/illtyped
  [(stlc-tc Γ c) (tyc c)]
  [(stlc-tc Γ (op t ..._1)) 
   B
   (where (A_1 ... -> B) (tyop op))
   (where (A_2 ...) ((stlc-tc Γ t) ...))
   #;(side-condition (equal? (term (A_1 ...)) (term (A_2 ...))))
   (side-condition (term (same-tys/Γ Γ (A_1 ...) (A_2 ...))))]
  [(stlc-tc (Γ x A) x)
   A]
  [(stlc-tc (Γ y A) x)
   (stlc-tc Γ x)]
  [(stlc-tc Γ (λ x : A t))
   ((close-ty Γ A) -> B)
   (where B (stlc-tc (Γ x A) t))] 
  [(stlc-tc Γ (t s))
   B
   (where (A_1 -> B) (stlc-tc Γ t))
   (where A_2 (stlc-tc Γ s))
   (side-condition (term (same-ty/Γ Γ A_1 A_2)))]
  [(stlc-tc Γ t) illtyped])
  
(define-metafunction stlc
  tyop : op -> (A ... -> B)
  [(tyop +) (Int Int -> Int)]
  [(tyop -) (Int Int -> Int)]
  [(tyop =) (Int Int -> Bool)]
  [(tyop if) (Bool -> ((Int -> Int) -> ((Int -> Int) -> Int)))])

(define-metafunction stlc
  tyc : c -> ι
  [(tyc number) Int]
  [(tyc false) Bool]
  [(tyc true) Bool])
