#lang racket
(require redex/reduction-semantics)
(provide subst)

(define-language empty-language)

(define-metafunction empty-language
  subst : variable any any -> any
  ;; 1. variable_1 bound, so don't continue in λ body
  [(subst variable_1 any_1 (λ variable_1 : any_2 any_3))
   (λ variable_1 : any_2 any_3)]
  [(subst variable_1 any_1 (Λ variable_1 any_2))
   (Λ variable_1 any_2)]
  [(subst variable_1 any_1 (ν variable_1 := any_2 any_3))
   (ν variable_1 := any_2 any_3)]
  [(subst variable_1 any_1 (∀ variable_1 any_2))
   (∀ variable_1 any_2)]
  
  ;; 2. general purpose capture avoiding cases
  [(subst variable_1 any_1 (λ variable_2 : any_2 any_3))
   (λ variable_new :
     (subst variable_1 
            any_1
            (subst-var variable_2 variable_new any_2))
     (subst variable_1 
            any_1
            (subst-var variable_2 variable_new any_3)))
   (where variable_new
          ,(variable-not-in
            (term (variable_1 any_1 any_2 any_3)) 
            (term variable_2)))]
  
  [(subst variable_1 any_1 (Λ variable_2 any_2))
   (Λ variable_new 
     (subst variable_1 
            any_1
            (subst-var variable_2 variable_new any_2)))
   (where variable_new
          ,(variable-not-in
            (term (variable_1 any_1 any_2)) 
            (term variable_2)))]
  
  [(subst variable_1 any_1 (ν variable_2 := any_2 any_3))
   (ν variable_new :=
      (subst variable_1 
             any_1
             (subst-var variable_2 variable_new any_2))
      (subst variable_1 
             any_1
             (subst-var variable_2 variable_new any_3)))
   (where variable_new
          ,(variable-not-in
            (term (variable_1 any_1 any_2 any_3)) 
            (term variable_2)))]
  
  [(subst variable_1 any_1 (∀ variable_2 any_2))
   (∀ variable_new 
     (subst variable_1 
            any_1
            (subst variable_2 variable_new any_2)))
   (where variable_new
          ,(variable-not-in
            (term (variable_1 any_1 any_2 any_3)) 
            (term variable_2)))]
  
  ;; 3. replace variable_1 with any_1
  [(subst variable_1 any_1 variable_1) any_1]
  ;; 4. variable_1 and variable_2 are different, so don't replace
  [(subst variable_1 any_1 variable_2) variable_2]
  ;; the last cases cover all other expressions
  [(subst variable_1 any_1 (any_2 ...))
   ((subst variable_1 any_1 any_2) ...)]
  [(subst variable_1 any_1 any_2) any_2])

(define-metafunction empty-language
  subst-var : variable variable any -> any
  [(subst-var variable_1 any_1 variable_1) any_1]
  [(subst-var variable_1 any_1 (any_2 ...)) 
   ((subst-var variable_1 any_1 any_2) ...)]
  [(subst-var variable_1 any_1 any_2) any_2])
